# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A constructor for building a minimal ontology, both for a pedagogical example and for
tests.
"""

from __future__ import annotations

from pyiron_ontology.constructor import Constructor

import owlready2 as owl


class ExampleOntology(Constructor):
    """
    The ontology represents simple three layer diagrams input->middle->output, with
    different choices available at each layer.
    The final output layer has four options, which have restrictions on all combinations
    in the (`generic`, `requirements`) square, allowing for progressively fewer
    possible three-layer graphs.

    In the middle layer there's a bit of extra unused input and output, just so we can
    test to make sure we count everything correctly in non-trivial ways.

    At the domain knowledge level (i.e. the `Generic` things we define) we show how to
    achieve mutual exclusion by use of disjoint classes and instantiating individuals.

    Terminology goes: I:Input:IM:Middle:MO:Output:O, with letters and numbers to
    distinguish siblings/inheritance.
    """

    def __init__(self, name: str = "example", closed: bool = True, strict: bool = True):
        super().__init__(name=name, closed=closed, strict=strict)

    def _make_specific_declarations(self):
        onto = self.onto
        with onto:

            class I(onto.Generic):
                pass

            class I1(I):
                pass

            class I2(I):
                pass

            owl.AllDisjoint([I1, I2])

            class IM(onto.Generic):
                pass

            class IMOptional(onto.Generic):
                pass

            class MO(onto.Generic):
                pass

            class MO1(MO):
                pass

            class MO2(MO):
                pass

            class MO2A(MO2):
                pass

            class MO2B(MO2):
                pass

            owl.AllDisjoint([MO1, MO2])

            class MOUnused(onto.Generic):
                pass

            class O(onto.Generic):
                pass

            owl.AllDisjoint([O, MO, MOUnused, IMOptional, IM, I])

        input1 = onto.Function("input1")
        input1_inp = onto.Input(
            name="input1_inp",
            mandatory_input_of=input1,
            generic=I1(),
        )
        input1_out = onto.Output(
            name="input1_out",
            output_of=input1,
            generic=IM(),
        )

        input2 = onto.Function("input2")
        input2_inp = onto.Input(
            name="input2_inp",
            mandatory_input_of=input2,
            generic=I2(),
        )
        input2_out = onto.Output(
            name="input2_out",
            output_of=input2,
            generic=IM(),
        )

        middle1 = onto.Function("middle1")
        middle1_inp1 = onto.Input(
            name="middle1_inp1",
            mandatory_input_of=middle1,
            generic=IM(),
            transitive_requirements=[I()],
        )
        middle1_inp2 = onto.Input(
            name="middle1_inp2",
            optional_input_of=middle1,
            generic=IMOptional(),
        )
        middle1_out1 = onto.Output(
            name="middle1_out1",
            output_of=middle1,
            generic=MO1(),
        )
        middle1_out2 = onto.Output(
            name="middle1_out2",
            output_of=middle1,
            generic=MOUnused(),
        )

        middle2 = onto.Function("middle2")
        middle2_inp1 = onto.Input(
            name="middle2_inp1",
            mandatory_input_of=middle2,
            generic=IM(),
            transitive_requirements=[I()],
        )
        middle2_inp2 = onto.Input(
            name="middle2_inp2",
            optional_input_of=middle2,
            generic=IMOptional(),
        )
        middle2_out1 = onto.Output(
            name="middle2_out1",
            output_of=middle2,
            generic=MO2B(),
        )
        middle2_out2 = onto.Output(
            name="middle2_out2",
            output_of=middle2,
            generic=MOUnused(),
        )

        output1 = onto.Function("output1")  # Allows all paths
        output1_inp = onto.Input(
            name="output1_inp",
            mandatory_input_of=output1,
            generic=MO(),
        )
        output1_out = onto.Output(
            name="output1_out",
            output_of=output1,
            generic=O(),
        )

        output2 = onto.Function("output2")  # Must pass through middle1
        output2_inp = onto.Input(
            name="output2_inp",
            mandatory_input_of=output2,
            generic=MO1(),
        )
        output2_out = onto.Output(
            name="output2_out",
            output_of=output2,
            generic=O(),
        )

        output3 = onto.Function("output3")  # Must end at input1
        output3_inp = onto.Input(
            name="output3_inp",
            mandatory_input_of=output3,
            generic=MO(),
            requirements=[I1()],
        )
        output3_out = onto.Output(
            name="output3_out",
            output_of=output3,
            generic=O(),
        )

        output4 = onto.Function("output4")  # Only the path middle2 -> input2
        output4_inp = onto.Input(
            name="output4_inp",
            mandatory_input_of=output4,
            generic=MO2(),  # Finds middle2_out1;
            # But it only finds _more specific_ output, if these generics are swapped,
            # it finds nothing.
            # Is this really desirable, or do we want anything
            requirements=[I2()],
        )
        output4_out = onto.Output(
            name="output4_out",
            output_of=output4,
            generic=O(),
        )

        # I don't yet test dual inheritance in this example!!!



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A constructor for building the atomistics ontology from python classes.
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    import pyiron_atomistics
    from pyiron_ontology.atomistics.constructor import AtomisticsOntology

    onto = AtomisticsOntology().onto


class AtomisticsReasoner:
    def __init__(self, ontology):
        self.onto = ontology

    @staticmethod
    def _get_ref_job(job):
        ref_job_name = list(set(job.project_hdf5.list_groups()) - {"input", "output"})
        if len(ref_job_name) == 1:
            return ref_job_name[0]

    def _get_job_type(self, job):
        ref_job_name = self._get_ref_job(job)
        if ref_job_name is not None:
            return job[f"{ref_job_name}/TYPE"].split(".")[-1][:-2]
        else:
            return job["TYPE"]

    @staticmethod
    def _alloy_sql(el):
        """Convert to SQL search string"""
        if el is None:
            return "%"
        return f"%{el}%"

    def search_database_for_property(
        self,
        my_property: onto.Generic,
        project: pyiron_atomistics.Project,
        select_alloy: Optional[str] = None,
    ):
        """
        Use the pyiron database to search for instances of an ontological generic
        parameter. Optionally filter by the chemistry of the job.

        Args:
            my_property:
            project:
            select_alloy:

        Returns:

        """
        pd_header = [
            "Chemical Formula",
            f"{my_property.__class__}",
            "unit",
            "Engine",
        ]
        pd_dic = {k: [] for k in pd_header}

        for specific_property in my_property.indirect_outputs:
            property_hdf_path = specific_property.hdf_path
            df_murn = pd.DataFrame(
                project.db.get_items_dict(
                    {
                        "hamilton": specific_property.output_of.pyiron_name,
                        "chemicalformula": self._alloy_sql(select_alloy),
                        "project": f"%{project.path}%",
                    }
                )
            )

            for _, row in df_murn.iterrows():
                job_hdf = project.inspect(row.id)
                pd_dic["Chemical Formula"].append(row.chemicalformula)
                output_property = job_hdf[property_hdf_path]
                value = output_property
                pd_dic[pd_header[1]].append(value)
                pd_dic["unit"].append(specific_property.unit)
                pd_dic["Engine"].append(self._get_job_type(job_hdf))
            return pd.DataFrame(pd_dic)



from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from pyiron_ontology.atomistics.constructor import AtomisticsOntology
from pyiron_ontology.constructor import Constructor
from pyiron_ontology.dynamic import DynamicOntologies as dynamic

from pyiron_ontology.atomistics.reasoning import AtomisticsReasoner



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A tree structure for ontologically-informed workflows
"""

from numpy import argsort


class NodeTree:
    def __init__(self, value, parent=None):
        self.value = value
        self.children = []
        self.parent = parent
        if parent is not None:
            parent.children.append(self)

    def render(self, depth=0, order_alphabetically=True):
        tabs = "".join(["  "] * depth)
        print(f"{tabs}{self.value.name}")
        children = (
            [self.children[n] for n in argsort([str(c.value) for c in self.children])]
            if order_alphabetically
            else self.children
        )
        for child in children:
            child.render(depth=depth + 1)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A parent class for the constructors of all pyiron ontologies.
"""

from __future__ import annotations

from typing import Optional
from warnings import warn

import owlready2 as owl
import pint

from pyiron_ontology.workflow import NodeTree

UREG = pint.UnitRegistry()


class Constructor:
    def __init__(
        self,
        name: str,
        closed: bool = True,
        strict: bool = False,
        debug: int = 0,
    ):
        onto = owl.get_ontology(f"file://{name}.owl")
        self.onto = onto
        self._make_universal_declarations()
        self._make_specific_declarations()
        # TODO: Introduce a "from_csv" option for constructing, and leverage
        #       `all_classes=False` in `declare_classes`?
        self.sync(closed=closed, strict=strict, debug=debug)

    def sync(
        self,
        closed=True,
        infer_property_values=True,
        infer_data_property_values=True,
        debug=0,
        strict=True,
    ):
        if closed:
            owl.close_world(self.onto.PyObject)
        with self.onto:
            owl.sync_reasoner_pellet(
                infer_property_values=infer_property_values,
                infer_data_property_values=infer_data_property_values,
                debug=debug,
            )
        inconsistent = list(self.onto.inconsistent_classes())
        if len(inconsistent) > 0:
            msg = f"Inconsistent classes were found in the ontology: {inconsistent}"
            if strict:
                raise RuntimeError(msg)
            else:
                warn(msg)

    def save(self):
        self.onto.save()

    def _make_specific_declarations(self):
        pass

    def _make_universal_declarations(self):
        with self.onto:

            class PyironOntoThing(owl.Thing):
                def get_sources(
                    self, additional_requirements: list[Generic] = None
                ) -> list[WorkflowThing]:
                    raise NotImplementedError

                def get_source_tree(self, additional_requirements=None):
                    return build_tree(
                        self, additional_requirements=additional_requirements
                    )

                def get_source_path(self, *path_indices: int):
                    return build_path(self, *path_indices)

            class Parameter(PyironOntoThing):
                def unit_conversion(self, other_unit: str) -> float:
                    if self.unit is not None:
                        return UREG(self.unit).to(other_unit).magnitude
                    else:
                        raise ValueError("Parameters must have a unit specified")

            class has_unit(Parameter >> str, owl.FunctionalProperty):
                class_property_type = ["some"]
                python_name = "unit"

            class Generic(Parameter):
                def get_sources(
                    self, additional_requirements: list[Generic] = None
                ) -> list[Output]:
                    return [
                        out
                        for out in self.indirect_outputs
                        if (
                            out.satisfies(additional_requirements)
                            if additional_requirements is not None
                            else True
                        )
                    ]

                @staticmethod
                def only_get_thing_classes(things):
                    return [
                        is_a_class
                        for is_a_class in things
                        if isinstance(is_a_class, owl.ThingClass)
                    ]

                @property
                def indirect_things(self):
                    return self.only_get_thing_classes(self.INDIRECT_is_a)

                @property
                def indirect_io(self) -> list[Parameter]:
                    generic_classes = self.only_get_thing_classes(self.is_a)
                    unique_instances = list(
                        set(generic_classes[0].instances()).union(
                            *[gc.instances() for gc in generic_classes[1:]]
                        )
                    )
                    return [p for ui in unique_instances for p in ui.parameters]

                @property
                def indirect_outputs(self) -> list[Output]:
                    return [p for p in self.indirect_io if Output in p.is_a]

                @property
                def indirect_disjoints_set(self) -> set[Generic]:
                    return get_disjoints_set(self.indirect_things)

                @property
                def representation_info(self):
                    """
                    A more computationally efficient call when you know you need both
                    the `indirect_disjoints` _and_ `indirect_things` properties at once.

                    Returns:
                        list, set: indirect things, indirect disjoints
                    """
                    indirect_things = self.indirect_things
                    indirect_disjoints = get_disjoints_set(indirect_things)
                    return indirect_things, indirect_disjoints

                @classmethod
                def class_is_indirectly_disjoint_with(cls, other: owl.ThingClass):
                    ancestors1 = list(cls.ancestors())
                    ancestors2 = list(other.ancestors())
                    combined_disjoints = get_disjoints_set(ancestors1).union(
                        get_disjoints_set(ancestors2)
                    )
                    combined_ancestors = set(ancestors1).union(ancestors2)
                    return len(combined_disjoints.intersection(combined_ancestors)) > 0

                def has_a_representation_among_others(self, others_info):
                    my_things, my_disjoints = self.representation_info
                    return any(
                        compatible_classes(
                            my_things,
                            my_disjoints,
                            other_things,
                            other_disjoints,
                        )
                        for (other_things, other_disjoints) in others_info
                    )

            class WorkflowThing(PyironOntoThing):
                pass

            class Function(WorkflowThing):
                def get_sources(
                    self, additional_requirements: list[Generic] = None
                ) -> list[Input]:
                    return self.mandatory_inputs

                @property
                def inputs(self):
                    return self.mandatory_inputs + self.optional_inputs

                @property
                def options(self):
                    return [
                        opt
                        for inp in self.inputs
                        for opt in [inp.generic]
                        + inp.requirements
                        + inp.transitive_requirements
                    ]

            class IO(Parameter, WorkflowThing):
                pass

            class has_generic(IO >> Generic, owl.FunctionalProperty):
                python_name = "generic"

            class has_for_parameter(Generic >> IO, owl.InverseFunctionalProperty):
                python_name = "parameters"
                inverse_property = has_generic

            class has_hdf_path(IO >> str, owl.FunctionalProperty):
                python_name = "hdf_path"

            class Output(IO):
                def get_sources(
                    self, additional_requirements: list[Generic] = None
                ) -> list[Function]:
                    return [self.output_of]

                @property
                def options(self):
                    return self.output_of.options

                def satisfies(self, requirements: list[Generic]) -> bool:
                    others_info = [
                        other.representation_info
                        for other in self.options + [self.generic]
                    ]
                    return all(
                        requirement.has_a_representation_among_others(others_info)
                        for requirement in requirements
                    )

            class is_output_of(Output >> Function, owl.FunctionalProperty):
                python_name = "output_of"

            class has_for_output(Function >> Output, owl.InverseFunctionalProperty):
                python_name = "outputs"
                inverse_property = is_output_of

            class Input(IO):
                def get_sources(
                    self, additional_requirements: Optional[list[Generic]] = None
                ) -> list[Output]:
                    return self.get_sources_and_passed_requirements(
                        additional_requirements=additional_requirements
                    )[0]

                def get_sources_and_passed_requirements(
                    self, additional_requirements: Optional[list[Generic]] = None
                ) -> tuple[list[Output], list[Generic]]:
                    requirements = self.get_requirements(
                        additional_requirements=additional_requirements
                    )
                    sources = self.generic.get_sources(
                        additional_requirements=requirements
                    )
                    return sources, requirements

                def get_requirements(self, additional_requirements=None):
                    """
                    For each additional requirement, see if it is as or more specific
                    than an existing requirement (from among the generic class,
                    requirements, and transitive requirements), and if so keep it
                    (discarding the original if in the generic class or requirements,
                    appending if it's a transitive requirement that we're actually
                    receiving).
                    """
                    if additional_requirements is None:
                        return [self.generic] + self.requirements
                    requirements = [self.generic] + self.requirements

                    base_infos = [other.representation_info for other in requirements]
                    transitive_infos = [
                        other.representation_info
                        for other in self.transitive_requirements
                    ]

                    for add_req in additional_requirements:
                        add_things, add_disjoints = add_req.representation_info
                        used = False  # For early breaking if we use the additional req
                        for i, (base_things, base_disjoints) in enumerate(base_infos):
                            if self.candidate_is_as_or_more_specific_than(
                                add_things, base_disjoints, base_things
                            ):
                                requirements[i] = add_req  # Overwrite the thing you're
                                # more specific than
                                used = True
                                break
                        if used:
                            continue

                        for trans_things, trans_disjoints in transitive_infos:
                            # If you haven't found the additional requirement yet,
                            # check if it's in the allowed transitive requirements
                            if compatible_classes(
                                add_things,
                                add_disjoints,
                                trans_things,
                                trans_disjoints,
                            ):
                                requirements.append(add_req)
                                break
                    return requirements

                @staticmethod
                def candidate_is_as_or_more_specific_than(
                    candidate_things, ref_disjoints, ref_things
                ) -> bool:
                    not_disjoint = (
                        len(ref_disjoints.intersection(candidate_things)) == 0
                    )
                    return not_disjoint and set(ref_things).issubset(candidate_things)

            class is_optional_input_of(Input >> Function, owl.FunctionalProperty):
                python_name = "optional_input_of"

            class has_for_optional_input(
                Function >> Input, owl.InverseFunctionalProperty
            ):
                python_name = "optional_inputs"
                inverse_property = is_optional_input_of

            class is_mandatory_input_of(Input >> Function, owl.FunctionalProperty):
                python_name = "mandatory_input_of"

            class has_for_mandatory_input(
                Function >> Input, owl.InverseFunctionalProperty
            ):
                python_name = "mandatory_inputs"
                inverse_property = is_mandatory_input_of

            class has_for_requirement(Input >> Generic):
                python_name = "requirements"

            class is_requirement_of(Generic >> Input):
                python_name = "requirement_of"
                inverse_property = has_for_requirement

            class has_for_transitive_requirement(Input >> Generic):
                python_name = "transitive_requirements"

            class is_transitive_requirement_of(Generic >> Input):
                python_name = "transitive_requirement_of"
                inverse_property = has_for_transitive_requirement

            owl.AllDisjoint([is_optional_input_of, is_mandatory_input_of])
            owl.AllDisjoint([Input, Function, Output, Generic])

        def compatible_classes(
            things1: list[owl.ThingClass],
            disjoints1: set[owl.ThingClass],
            things2: list[owl.ThingClass],
            disjoints2: set[owl.ThingClass],
        ):
            """
            Given the `is_a` and disjoint classes of two individuals, checks whether
            they are compatible -- i.e. whether the classes of one are in the disjoints
            of the other (which would lead to incompatibility).

            Args:
                things1 (list[owl.ThingClass]): Classes of the first indivual.
                disjoints1 (set[owl.ThingClass]): Disjoints of the first individual.
                things2 (list[owl.ThingClass]): Classes of the second individual.
                disjoints2 (set[owl.ThingClass]):

            Returns:
                (bool): Whether any classes of one individual are in the disjoints of
                    the other.
            """
            # Put 0 first so we can skip the second evaluation when the first fails
            return (
                0
                == len(disjoints1.intersection(things2))
                == len(disjoints2.intersection(things1))
            )

        def get_disjoints_set(classes: list[owl.ThingClass]):
            """
            For a list of things, get the set of all the things they're disjoint
            to
            """
            disjoints = []
            for thing in classes:
                if thing == owl.Thing:
                    continue
                try:
                    entities = list(next(thing.disjoints()).entities)
                    # The entities are the actual classes that are disjoint
                    # The entities for each of the disjoints are ideantical,
                    # so we can just use `next` to grab the first one
                    entities.remove(thing)
                    # The entities of our disjoint include us, so remove us
                    disjoints += entities
                except StopIteration:
                    # If the disjoints are empty, just continue
                    continue
            return set(disjoints)

        def build_tree(
            parameter, parent=None, additional_requirements=None
        ) -> NodeTree:
            node = NodeTree(parameter, parent=parent)

            if isinstance(parameter, Input):
                (
                    sources,
                    additional_requirements,
                ) = parameter.get_sources_and_passed_requirements(
                    additional_requirements=additional_requirements
                )  # Snag the accepted transitive requirements as well
            else:
                sources = parameter.get_sources(
                    additional_requirements=additional_requirements
                )

            for source in sources:
                build_tree(
                    source, parent=node, additional_requirements=additional_requirements
                )

            return node

        def build_path(
            parameter, *path_indices: int, parent=None, additional_requirements=None
        ):
            node = NodeTree(parameter, parent=parent)

            sources = parameter.get_sources(
                additional_requirements=additional_requirements
            )

            if len(path_indices) > 0:
                i, path_indices = path_indices[0], path_indices[1:]
                source = sources[i]
                _, sources = build_path(
                    source,
                    *path_indices,
                    parent=node,
                    additional_requirements=additional_requirements,
                )

            return node, sources



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A constructor for building the atomistics ontology from python classes.
"""

from __future__ import annotations

from pyiron_ontology.constructor import Constructor

import owlready2 as owl


class AtomisticsOntology(Constructor):
    def __init__(
        self, name: str = "atomistics", closed: bool = True, strict: bool = False
    ):
        super().__init__(name=name, closed=closed, strict=strict)

    def _make_specific_declarations(self):
        Generic = self.onto.Generic
        Input = self.onto.Input
        Output = self.onto.Output

        with self.onto:

            class AtomisticsFunction(self.onto.Function):
                pass

            class has_pyiron_name(AtomisticsFunction >> str, owl.FunctionalProperty):
                python_name = "pyiron_name"

            Function = AtomisticsFunction

            class UserInput(Generic):
                pass

            class PyironObject(Generic):
                pass

            class PhysicalProperty(Generic):
                pass  # Add units, etc

            owl.AllDisjoint([PyironObject, PhysicalProperty])

            class Energy(PhysicalProperty):
                pass

            class Force(PhysicalProperty):
                pass

            class ChemicalElement(PhysicalProperty):
                pass

            owl.AllDisjoint([Energy, Force, ChemicalElement])

            class MaterialProperty(PhysicalProperty):
                pass

            class BulkModulus(MaterialProperty):
                pass

            class BPrime(MaterialProperty):
                pass

            class SurfaceEnergy(MaterialProperty):
                pass

            owl.AllDisjoint([BulkModulus, BPrime, SurfaceEnergy])

            class Dimensional(Generic):
                pass

            class OneD(Dimensional):
                pass

            class TwoD(Dimensional):
                pass

            class ThreeD(Dimensional):
                pass

            owl.AllDisjoint([OneD, TwoD, ThreeD])

            class Structure(PyironObject, Dimensional):
                pass

            class Defected(Structure):
                pass

            class HasDislocation(Defected):
                pass

            class HasVacancy(Defected):
                pass

            class HasInterface(Defected):
                pass

            class HasGB(HasInterface):
                pass

            class HasSurface(HasInterface):
                pass

            class HasPB(HasInterface):
                pass

            class Bulk(Structure):
                equivalent_to = [Structure & owl.Not(Defected)]

            # equivalent_to = [Structure & owl.Not(Defected)]
            owl.AllDisjoint([Bulk, Defected])  # Not even needed given Bulk definition
            owl.AllDisjoint([OneD, HasGB])
            owl.AllDisjoint([OneD, HasDislocation])

            class PyironProject(PyironObject):
                pass

            class AtomisticsProject(PyironProject):
                pass

            class PyironJob(PyironObject):
                pass

            class AtomisticsJob(PyironJob):
                pass

            class Lammps(AtomisticsJob):
                pass

            class Vasp(AtomisticsJob):
                pass

            owl.AllDisjoint([Structure, PyironProject, PyironJob])

            project = Function(name="project")
            project_input_name = Input(
                optional_input_of=project,
                name=f"{project.name}_input_name",
                generic=UserInput(),
            )
            project_output_atomistics_project = Output(
                output_of=project,
                name=f"{project.name}_output_atomistics_project",
                generic=AtomisticsProject(),
            )

            bulk_structure = Function(name="bulk_structure")
            bulk_structure_input_element = Input(
                optional_input_of=bulk_structure,
                name=f"{bulk_structure.name}_input_element",
                generic=Generic(is_a=[ChemicalElement, UserInput]),
                hdf_path="input/element",
            )
            bulk_structure_output_structure = Output(
                output_of=bulk_structure,
                name=f"{bulk_structure.name}_output_structure",
                generic=Structure(is_a=[Bulk, ThreeD]),
                hdf_path="output/structure",
            )

            surface_structure = Function("surface_structure")
            surface_structure_input_element = Input(
                optional_input_of=surface_structure,
                name=f"{surface_structure.name}_input_element",
                generic=Generic(is_a=[ChemicalElement, UserInput]),
                hdf_path="input/element",
            )
            surface_structure_output_structure = Output(
                output_of=surface_structure,
                name=f"{surface_structure.name}_output_structure",
                generic=Structure(is_a=[HasSurface, ThreeD]),
                hdf_path="output/structure",
            )

            lammps = Function("lammps", pyiron_name="Lammps")
            lammps_input_project = Input(
                name=f"{lammps.name}_input_project",
                generic=AtomisticsProject(),
                mandatory_input_of=lammps,
            )
            lammps_input_structure = Input(
                mandatory_input_of=lammps,
                name=f"{lammps.name}_input_structure",
                generic=Structure(),
                hdf_path="input/structure",
            )
            lammps_output_job = Output(
                output_of=lammps, name=f"{lammps.name}_output_job", generic=Lammps()
            )

            vasp = Function("vasp", pyiron_name="Vasp")
            vasp_input_project = Input(
                name=f"{vasp.name}_input_project",
                generic=AtomisticsProject(),
                mandatory_input_of=vasp,
            )
            vasp_input_structure = Input(
                mandatory_input_of=vasp,
                name=f"{vasp.name}_input_structure",
                generic=Generic(is_a=[Structure, ThreeD]),
                # Can't be onto.Structure(is_a=[onto.ThreeD]) because is_a _overrides_ the
                # instantiated class, and ThreeD is not a child of Structure!
                hdf_path="input/structure",
            )
            vasp_output_job = Output(
                output_of=vasp, name=f"{vasp.name}_output_job", generic=Vasp()
            )

            atomistic_taker = Function("atomistic_taker")
            atomistic_taker_job = Input(
                name="atomistic_taker_job",
                generic=AtomisticsJob(),
                mandatory_input_of=atomistic_taker,
                transitive_requirements=[Structure()],
            )
            atomistic_taker_output_energy_pot = Output(
                name="atomistic_taker_output_energy_pot",
                generic=Energy(),
                output_of=atomistic_taker,
            )
            atomistic_taker_output_forces = Output(
                name="atomistic_taker_output_forces",
                generic=Force(),
                output_of=atomistic_taker,
            )

            murnaghan = Function("murnaghan", pyiron_name="Murnaghan")
            murnaghan_input_project = Input(
                name=f"{murnaghan.name}_input_project",
                generic=AtomisticsProject(),
                mandatory_input_of=murnaghan,
            )
            murnaghan_input_job = Input(
                name=f"{murnaghan.name}_input_job",
                generic=AtomisticsJob(),
                mandatory_input_of=murnaghan,
                requirements=[Structure(is_a=[Bulk, ThreeD])],
                hdf_path="ref_job",
            )
            murnaghan_output_bulk_modulus = Output(
                name=f"{murnaghan.name}_output_bulk_modulus",
                generic=BulkModulus(),
                output_of=murnaghan,
                hdf_path="output/equilibrium_bulk_modulus",
                unit="GPa",
            )
            murnaghan_output_b_prime = Output(
                name=f"{murnaghan.name}_output_b_prime",
                generic=BPrime(),
                output_of=murnaghan,
                hdf_path="output/equilibrium_b_prime",
            )

            surface_energy = Function("surface_energy")
            surface_energy_input_bulk_structure = Input(
                name="surface_energy_input_bulk_structure",
                generic=Structure(is_a=[Bulk, ThreeD]),
                mandatory_input_of=surface_energy,
            )
            surface_energy_input_bulk_energy = Input(
                name="surface_energy_input_bulk_energy",
                generic=Energy(),
                mandatory_input_of=surface_energy,
                requirements=[Structure(is_a=[Bulk, ThreeD])],
            )
            surface_energy_input_slab_structure = Input(
                name="surface_energy_input_slab_structure",
                generic=Structure(is_a=[HasSurface, ThreeD]),
                mandatory_input_of=surface_energy,
            )
            surface_energy_input_slab_energy = Input(
                name="surface_energy_input_slab_energy",
                generic=Energy(),
                mandatory_input_of=surface_energy,
                requirements=[Structure(is_a=[HasSurface, ThreeD])],
            )
            surface_energy_output_surface_energy = Output(
                name=f"{surface_energy.name}_output_surface_energy",
                generic=SurfaceEnergy(),
                output_of=surface_energy,
            )



"""
Setuptools based setup module
"""
from setuptools import setup, find_packages
import versioneer

setup(
    name='pyiron_ontology',
    version=versioneer.get_version(),
    description='pyiron_ontology - module extension to pyiron.',
    long_description='http://pyiron.org',

    url='https://github.com/pyiron/pyiron_ontology',
    author='Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department',
    author_email='liamhuber@greyhavensolutions.com',
    license='BSD',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],

    keywords='pyiron',
    packages=find_packages(exclude=["*tests*", "*docs*", "*binder*", "*conda*", "*notebooks*", "*.ci_support*"]),
    install_requires=[
        'numpy>=1.26.3',
        'owlready2>=0.45',
        'pandas>=2.1.4',
        'pint>=0.23',
    ],
    cmdclass=versioneer.get_cmdclass(),

    )



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A class for lazy construction of the ontologies
"""

from pyiron_ontology.atomistics.constructor import AtomisticsOntology
from pyiron_ontology.example.constructor import ExampleOntology


class DynamicOntologies:
    _atomistics = None
    _example = None

    @classmethod
    def atomistics(cls):
        if cls._atomistics is None:
            cls._atomistics = AtomisticsOntology().onto
        return cls._atomistics

    @classmethod
    def example(cls):
        if cls._example is None:
            cls._example = ExampleOntology().onto
        return cls._example
